==> p6-0.4.0.tgz: package/src/cstore.js <==
import * as ctypes from './ctypes';
export default function ColumnStore(option){
    var cstore   = (this instanceof ColumnStore) ? this : {},
        columns  = [],                  // column-based binary data
        size     = option.size  || 0,   // max size
        count    = option.count || 0,   // number of entries stored
        types    = option.types || [],  // types of the columns
        attributes = option.attributes || option.keys || option.names || [],  // column attributes
        struct   = option.struct|| option.schema || {},
        CAMs     = option.CAMs  || {},  // content access memory
        TLBs     = option.TLBs  || {},  // table lookaside buffer
        colStats = {},
        colAlloc = {},
        colRead  = {},                  // functions for reading values
        skip     = option.skip  || 0;

    if(option.struct) initStruct(option.struct);

    function initCStore() {
        if(size && types.length === attributes.length && types.length > 0) {
            attributes.forEach(function(c, i){
                configureColumn(i);
                columns[i] = new colAlloc[c](size);
                if(!columns.hasOwnProperty(c))
                    Object.defineProperty(columns, c, {
                        get: function() { return columns[i]; }
                    });
            });
            columns.attributes = attributes;
            columns.keys = attributes;
            columns.types = types;
            columns.struct = struct;
            columns.TLBs = TLBs;
            columns.CAMs = CAMs;
            columns.size = size;
            columns.get = function(c) {
                var index = attributes.indexOf(c);
                if(index < 0 ) throw new Error("Error: No column named " + c);
                return columns[index];
            }
        }
        return cstore;
    }

    function initStruct(s) {
        struct = s;
        if(Array.isArray(struct)) {
            struct.forEach(function(s){
                attributes.push(s.name);
                types.push(s.type);
            })
        } else {
            for(var k in struct){
                attributes.push(k);
                types.push(struct[k]);
            }
        }
        return initCStore();
    }

    function configureColumn(cid) {
        if(typeof(cid) == "string") cid = attributes.indexOf(cid);
        var f = attributes[cid];
        colAlloc[f] = ctypes[types[cid]];

        if(colAlloc[f] === ctypes.string){
            TLBs[f] = [];
            CAMs[f] = {};
            colRead[f] = function(value) {
                if(!CAMs[f].hasOwnProperty(value)){
                    CAMs[f][value] = TLBs[f].length;
                    TLBs[f].push(value);
                }
                return CAMs[f][value];
            };
        } else if(
            colAlloc[f] === ctypes.int ||
            colAlloc[f] === ctypes.short ||
            colAlloc[f] === ctypes.integer
        ) {
            colRead[f] = function(value) {  return parseInt(value) || 0; };
        } else if(
            colAlloc[f] === ctypes.float ||
            colAlloc[f] === ctypes.double ||
            colAlloc[f] === ctypes.numeric
        ){
            colRead[f] = function(value) {  return parseFloat(value) || 0.0; };
        } else if(
                colAlloc[f] === ctypes.time ||
                colAlloc[f] === ctypes.temporal
        ) {
            colRead[f] = function(value) {  return parseFloat(value) || 0.0; };
        } else {
            throw new Error("Invalid data type for TypedArray data!")
        }
    }

    cstore.addRows = function(rowArray) {
        if(count === 0 && skip > 0) {
            for(var j = 0; j<skip; j++)
                rowArray.shift();
        }
        rowArray.forEach(function(row, i){
            row.forEach(function(v,j){
                columns[j][count] = colRead[attributes[j]](v);
            });
            count++;
        });
        return count;
    }

    cstore.addColumn = function(arg) {
        var props = arg || {},
            columnData = props.data || props.array,
            columnName = props.name,
            columnType = props.dtype,
            values = props.values || [];

        var cid = attributes.indexOf(columnName);
        if( cid < 0) {
            attributes.push(columnName);
            types.push(columnType);
            configureColumn(columnName);
            cid = types.length - 1;
            Object.defineProperty(columns, columnName, {
                get: function() { return columns[cid]; }
            });
        }

        if(columnData instanceof ctypes[types[cid]]) {
            columns[cid] = columnData;
            if(values.length) {
                TLBs[columnName] = values;
                CAMs[columnName] = {};
                values.forEach(function(value, vi){
                    CAMs[columnName][value] = vi;
                })
            }
        } else if(ArrayBuffer.isView(columnData)){
            columns[cid] = new colAlloc[columnName](size);
            for(var di = 0; di < size; di++) {
                columns[cid][di] = colRead[columnName](columnData[di]);
            }
        } else {
            throw new Error("Error: Invalid data type for columnArray!");
        }
        size = count = columnData.length;
    }

    cstore.metadata = cstore.info = function() {
        return {
            size: size,
            count: count,
            attributes: attributes,
            types: types,
            TLBs: TLBs,
            CAMs: CAMs,
            stats: cstore.stats()
        }
    }

    cstore.data = cstore.columns = function() {
        return columns;
    }

    cstore.stats = function(col){
        var col = col || attributes;
        col.forEach(function(name, c){
            if(!colStats[c]){
                var min, max, avg;
                min = max = avg = columns[c][0];

                for(var i = 1; i < columns[c].length; i++){
                    var d = columns[c][i];
                    if(d > max) max = d;
                    else if(d < min) min = d;
                    avg = avg - (avg-d) / i;
                }
                if(max == min) max += 0.000001;
                colStats[name] = {min: min, max: max, avg: avg};
            }
        })
        return colStats;
    }

    cstore.domains = function(col){
        var col = col || attributes,
            domains = [];

        col.forEach(function(name, c){
            domains[name] = [colStats[name].min, colStats[name].max];
        })
        return domains;
    }

    cstore.ctypes = function() {
        return ctypes;
    }

    cstore.size = size;

    return initCStore();
}

==> p6-0.5.4.tgz: package/src/cstore.js <==
import * as ctypes from './ctypes';
export default function ColumnStore(arg){
    var cstore   = (this instanceof ColumnStore) ? this : {},
        options = arg || {},
        columns  = [],                  // column-based binary data
        size     = options.size  || 0,   // max size
        count    = options.count || 0,   // number of entries stored
        types    = options.types || [],  // types of the columns
        attributes = options.attributes || options.keys || options.names || [],  // column attributes
        struct   = options.struct|| options.schema || {},
        strHashes     = options.strHashes  || {},  // content access memory
        strLists     = options.strLists  || {},  // table lookaside buffer
        colStats = {},
        colAlloc = {},
        colRead  = {},                  // functions for reading values
        skip     = options.skip  || 0;

    if(options.struct) initStruct(options.struct);

    function initCStore() {
        if(size && types.length === attributes.length && types.length > 0) {
            attributes.forEach(function(c, i){
                configureColumn(i);
                columns[i] = new colAlloc[c](size);
                if(!columns.hasOwnProperty(c))
                    Object.defineProperty(columns, c, {
                        get: function() { return columns[i]; }
                    });
            });
            columns.attributes = attributes;
            columns.keys = attributes;
            columns.types = types;
            columns.struct = struct;
            columns.strLists = strLists;
            columns.strHashes = strHashes;
            columns.size = size;
            columns.get = function(c) {
                var index = attributes.indexOf(c);
                if(index < 0 ) throw new Error("Error: No column named " + c);
                return columns[index];
            }
        } 
        return cstore;
    }

    function initStruct(s) {
        struct = s;
        if(Array.isArray(struct)) {
            struct.forEach(function(s){
                attributes.push(s.name);
                types.push(s.type);
            })
        } else {
            for(var k in struct){
                attributes.push(k);
                types.push(struct[k]);
            }
        }
        return struct;
    }

    function configureColumn(cid) {
        if(typeof(cid) == "string") cid = attributes.indexOf(cid);
        var f = attributes[cid];
        colAlloc[f] = ctypes[types[cid]];

        if(colAlloc[f] === ctypes.string){
            strLists[f] = [];
            strHashes[f] = {};
            colRead[f] = function(value) {
                if(!strHashes[f].hasOwnProperty(value)){
                    strHashes[f][value] = strLists[f].length;
                    strLists[f].push(value);
                }
                return strHashes[f][value];
            };
        } else if(
            colAlloc[f] === ctypes.int ||
            colAlloc[f] === ctypes.short ||
            colAlloc[f] === ctypes.integer
        ) {
            colRead[f] = function(value) {  return parseInt(value) || 0; };
        } else if(
            colAlloc[f] === ctypes.float ||
            colAlloc[f] === ctypes.double ||
            colAlloc[f] === ctypes.numeric
        ){
            colRead[f] = function(value) {  return parseFloat(value) || 0.0; };
        } else if(
                colAlloc[f] === ctypes.time ||
                colAlloc[f] === ctypes.temporal
        ) {
            colRead[f] = function(value) {  return parseFloat(value) || 0.0; };
        } else {
            throw new Error("Invalid data type for TypedArray data!")
        }
    }

    cstore.addRows = function(rowArray) {
        if(count === 0 && skip > 0) {
            for(var j = 0; j<skip; j++)
                rowArray.shift();
        }
        rowArray.forEach(function(row, i){
            row.forEach(function(v,j){
                columns[j][count] = colRead[attributes[j]](v);
            });
            count++;
        });
        return count;
    }

    cstore.addObjects = function(objArray) {
        if(count === 0 && skip > 0) {
            for(var j = 0; j<skip; j++)
                objArray.shift();
        }
        objArray.forEach(function(obj, i){
            Object.keys(obj).forEach(function(v,j){
                columns[j][count] = colRead[attributes[j]](obj[v]);
            });
            count++;
        });
        return count;
    }


    cstore.addColumn = function(arg) {
        var props = arg || {},
            columnData = props.data || props.array,
            columnName = props.name,
            columnType = props.dtype,
            values = props.values || [];

        var cid = attributes.indexOf(columnName);
        if( cid < 0) {
            attributes.push(columnName);
            types.push(columnType);
            configureColumn(columnName);
            cid = types.length - 1;
            Object.defineProperty(columns, columnName, {
                get: function() { return columns[cid]; }
            });
        }

        if(columnData instanceof ctypes[types[cid]]) {
            columns[cid] = columnData;
            if(values.length) {
                strLists[columnName] = values;
                strHashes[columnName] = {};
                values.forEach(function(value, vi){
                    strHashes[columnName][value] = vi;
                })
            }
        } else if(ArrayBuffer.isView(columnData)){
            columns[cid] = new colAlloc[columnName](size);
            for(var di = 0; di < size; di++) {
                columns[cid][di] = colRead[columnName](columnData[di]);
            }
        } else {
            throw new Error("Error: Invalid data type for columnArray!");
        }
        size = count = columnData.length;
    }

    cstore.metadata = cstore.info = function() {
        return {
            size: size,
            count: count,
            attributes: attributes,
            types: types,
            strLists: strLists,
            strHashes: strHashes,
            stats: cstore.stats()
        }
    }

    cstore.columns = function() {
        return columns;
    }

    cstore.data = function() {
        var data = columns;
        data.stats = cstore.stats();
        data.keys = attributes;
        data.size = size;
        data.strHashes = strHashes;
        data.strLists = strLists;
        data.dtypes = types;
        return data;
    }

    cstore.stats = function(col){
        var col = col || attributes;
        col.forEach(function(name, c){
            if(!colStats[c]){
                var min, max, avg;
                min = max = avg = columns[c][0];

                for(var i = 1; i < columns[c].length; i++){
                    var d = columns[c][i];
                    if(d > max) max = d;
                    else if(d < min) min = d;
                    avg = avg - (avg-d) / i;
                }
                if(max == min) max += 0.000001;
                colStats[name] = {min: min, max: max, avg: avg};
            }
        })
        return colStats;
    }

    cstore.domains = function(col){
        var col = col || attributes,
            domains = [];

        col.forEach(function(name, c){
            domains[name] = [colStats[name].min, colStats[name].max];
        })
        return domains;
    }

    cstore.ctypes = function() {
        return ctypes;
    }

    cstore.size = size;

    cstore.exportAsJSON = function() {
        var rows = new Array(size);
        for(var ri = 0; ri < size; ri++) {
            var dataFrame = {};
            attributes.forEach(function(attr, ai) {
                if(types[ai] == 'string') {
                    dataFrame[attr] = strLists[attr][columns[ai][ri]];
                } else {
                    dataFrame[attr] = columns[ai][ri];
                }
            })
            rows[ri] = dataFrame;
        }
        return rows;
    }

    cstore.exportAsRowArray = function() {
        var rows = new Array(size);
        for(var ri = 0; ri < size; ri++) {
            var row = new Array(attributes.length);
            attributes.forEach(function(attr, ai) {
                if(types[ai] == 'string') {
                    row[ai] = strLists[attr][columns[ai][ri]];
                } else {
                    row[ai] = columns[ai][ri];
                }
            })
            rows[ri] = row;
        }
        return rows;
    }

    cstore.export = function(arg) {
        var format = arg || 'json';
        if(format == 'rowArray') {
            return cstore.exportAsRowArray();
        } else {
            return cstore.exportAsJSON();
        }
    }

    cstore.import = function(arg) {
        var data = arg.data || [],
            schema = arg.schema || {};
        size = data.length;
        initStruct(schema);
        initCStore();
        cstore.addObjects(data);
    }

    return initCStore();
}
